#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <setjmp.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/times.h>
#include <math.h>
#include <memory.h>

#include "lint.h"
#include "config.h"
#include "interpret.h"
#include "object.h"
#include "exec.h"
#include "comm.h"
#include "mudstat.h"

jmp_buf error_recovery_context;
int error_recovery_context_exists = 0;
int change_memory_limit = 0;
/*
 * The 'current_time' is updated at every heart beat.
 */
int current_time;

static void cycle_hb_list (void);
extern struct object *command_giver, *current_interactive, *obj_list_destruct;
extern int num_player, d_flag, s_flag;
extern struct object *previous_ob, *master_ob;

struct object *current_heart_beat;

void call_heart_beat(), catch_alarm(int);
void load_first_objects(), prepare_ipc(),
    shutdowngame(), ed_cmd (char *),
    print_prompt(), call_out(),
    destruct2 (struct object *);

extern int get_message (char *, int), player_parser (char *),
    call_function_interactive (struct interactive *, char *),
    resort_free_list(), swap (struct object *);

extern void flush_all_player_mess();

extern int t_flag;
volatile int comm_time_to_call_heart_beat = 0; /* this is set by interrupt, */
	/* comm sets time_to_call_heart_beat sometime after */

/*
 * There are global variables that must be zeroed before any execution.
 * In case of errors, there will be a longjmp(), and the variables will
 * have to be cleared explicitely. They are normally maintained by the
 * code that use them.
 *
 * This routine must only be called from top level, not from inside
 * stack machine execution (as stack will be cleared).
 */
void 
clear_state() 
{
    extern struct object *previous_ob;

    current_object = 0;
    command_giver = 0;
    current_interactive = 0;
    previous_ob = 0;
    current_prog = 0;
    error_recovery_context_exists = 1;
    reset_machine(0);	/* Pop down the stack. */
#ifdef RUSAGE
    clear_cpu_stack();  /* gec or the stack would overfill some day */
                        /* error (setjmp) recovery always calls this ... */
#endif
}

void 
logon(struct object *ob)
{
    struct svalue *ret;
    struct object *save = current_object;

    /*
     * current_object must be set here, so that the static "logon" in
     * player.c can be called.
     */
    current_object = ob;
    ret = apply("logon", ob, 0, 0);
    if (ret == 0) {
	add_message("prog %s:\n", ob->name);
	fatal("Could not find logon on the player %s\n", ob->name);
    }
    current_object = save;
}

/*
 * Take a player command and parse it.
 * The command can also come from a NPC.
 * Beware that 'str' can be modified and extended !
 */
int 
parse_command(char *str, struct object *ob)
{
    struct object *save = command_giver;
    int res;

    command_giver = ob;
    res = player_parser(str);
    command_giver = save;
    return res;
}



/*
 * Temporary malloc, for memory allocated during one run in the backend loop
 * NOTE
 *   Memory allocated with tmpalloc() MUST be temprary because it is freed
 *   at the beginning of the backend() loop.
 *
 * Thanks to Marcus J Ranum (mjr@decuac.DEC.COM) for the idea.
 */
static 	char	**tmpmem = 0;
static 	int	tmp_size = 0, tmp_cur = 0;

#define TMPMEM_CHUNK 1024

void 
tmpclean()
{
    int il;

    for (il = tmp_cur - 1; il >=0; il--) /* Reversed order is probably good */
	free(tmpmem[il]);
    tmp_cur = 0;
}

char *
tmpalloc(int size)
{
    char  **list;

    if (tmp_cur >= tmp_size)
    {
	list = (char **)xalloc((tmp_size + TMPMEM_CHUNK) * sizeof(char *));

	if (tmp_size > 0)
	{
	    memcpy(list, tmpmem, tmp_size * sizeof(char *));
	    free((char *)tmpmem);
	}

	tmpmem = list;
	tmp_size += TMPMEM_CHUNK;
    }
    return tmpmem[tmp_cur++] = xalloc(size);
}

void 
tmpfree(char *p)
{
    return;
}
	

/*
 * This is the backend. We will stay here for ever (almost).
 */
int eval_cost;
int volatile interupted = 0;
int all_swapped;
static int number_swapped = 0;
void 
backend()
{
    char buff[2000];
    extern int game_is_being_shut_down;
    extern int slow_shut_down_to_do;
    extern void catch_io(int), startshutdowngame(int);
    extern int tot_alloc_object, tot_alloc_dest_object;
    read_snoop_file();
    (void) printf("Setting up ipc.\n");
    fflush(stdout);
    prepare_ipc();
    (void) signal(SIGHUP, startshutdowngame);
    (void) signal(SIGIO, catch_io);
    if (!setjmp(error_recovery_context))
    {
	if (!t_flag)
	    call_heart_beat();
    }
    /*
     * We come here after errors, and have to clear some global variables.
     */
    while(1) 
    {
	extern long time_to_swap;

	/* 
	 * Moved clear_state() inside the while loop after bugreport
	 * from Desmodus (TMI), 920112.
	 *
         * The call of clear_state() should not really have to be done
         * once every loop. However, there seem to be holes where the
         * state is not consistent. If these holes are removed,
         * then the call of clear_state() can be moved to just before the
         * while() - statment. *sigh* /Lars
	 */
	clear_state();
	tmpclean(); 		/* Free all temporary memory */
	eval_cost = 0;
	interupted = 0;
	remove_destructed_objects(); /* marion - before ref checks! */

        /*
         * Check if the memory limit has changed, this is an asynchronous
         * call, the reason for this coding.
         */
        if (change_memory_limit != 0)
        {
            if (change_memory_limit == 1)
                push_number(1);
            else
                push_number(0);
            (void) apply_master_ob("memory_reconfigure", 1);
            change_memory_limit = 0;
        }

#ifdef DEBUG
	if (d_flag & DEBUG_CHK_REF)
	    check_a_lot_ref_counts(0);
#endif
	if (game_is_being_shut_down)
	    shutdowngame();
	if (slow_shut_down_to_do)
	{
	    int tmp = slow_shut_down_to_do;
	    slow_shut_down_to_do = 0;
	    slow_shut_down(tmp);
	}

	if (comm_time_to_call_heart_beat)
	    call_heart_beat();
	
#if TIME_TO_SWAP
	if (current_time > all_swapped)
	    look_for_objects_to_swap((tot_alloc_object - tot_alloc_dest_object) * 
				 (current_time - all_swapped) / time_to_swap);
#endif
	if (get_message(buff, sizeof buff))
	{
	    void update_load_av (void);

	    update_load_av();
	    /*
	     * Now we have a string from the player. This string can go to
	     * one of several places. If it is prepended with a '!', then
	     * it is an escape from the 'ed' editor, so we send it
	     * as a command to the parser.
	     * If any object function is waiting for an input string, then
	     * send it there.
	     * Otherwise, send the string to the parser.
	     * The player_parser() will find that current_object is 0, and
	     * then set current_object to point to the object that defines
	     * the command. This will enable such functions to be static.
	     */
	    current_object = 0;
	    current_interactive = command_giver;
	    previous_ob = command_giver;

#ifdef DEBUG
	    if (!command_giver->interactive)
		fatal("Non interactive player in main loop !\n");
#endif
	    if (s_flag)
		reset_mudstatus();
	    if (buff[0] == '!' && command_giver->super)
		parse_command(buff + 1, command_giver);
	    else if (command_giver->interactive->ed_buffer)
		ed_cmd(buff);
	    else if (call_function_interactive(command_giver->interactive,
					       buff))
		;	/* Do nothing ! */
	    else
		parse_command(buff, command_giver);
	    /*
	     * Print a prompt if player is still here.
	     */
	    if (command_giver && command_giver->interactive)
	    {
		print_prompt();
		if (s_flag)
		    print_mudstatus(command_giver->name, eval_cost, 
				    get_millitime(), get_processtime());
	    }
	    flush_all_player_mess();
	    continue;
	}
#if TIME_TO_SWAP
	look_for_objects_to_swap(0);
#endif
	if (interupted)
	    continue;
	pause();
    }
}

/*
 * sig_alert is used to control the amount of memory that is
 * available. They check with the locally defined variables in the mudlib
 * for the space to set.  
 */
#if defined(_SEQUENT_)
void 
sig_alert(sig)
    int sig;
#else
void 
sig_alert(sig, code, scp, addr)
    int sig, code;
    struct sigcontext *scp;
    char *addr;
#endif
{
    switch(sig)
    {
    case SIGUSR1:
        change_memory_limit = -1;
        break;

    case SIGUSR2:
        change_memory_limit = 1;
        break;

    default:
        break;
    }
}

/*
 * Despite the name, this routine takes care of several things.
 * It will loop through all objects once every 10 minutes.
 *
 * If an object is found in a state of not having done reset, and the
 * delay to next reset has passed, then reset() will be done.
 *
 * If the object has a existed more than the time limit given for swapping,
 * then 'clean_up' will first be called in the object, after which it will
 * be swapped out if it still exists.
 *
 * There are some problems if the object self-destructs in clean_up, so
 * special care has to be taken of how the linked list is used.
 */
static void 
look_for_objects_to_reset() 
{
    jmp_buf save_error_recovery_context;
    int save_rec_exists;
    static struct object *ob;

    memcpy((char *) save_error_recovery_context,
	   (char *) error_recovery_context, sizeof error_recovery_context);
    save_rec_exists = error_recovery_context_exists;
    error_recovery_context_exists = 1;

    /*
     * Should this object have reset(1) called ?
     */
    while(obj_list->next_reset <= current_time)
    {
	ob = obj_list;
        if (setjmp(error_recovery_context)) 		/* amylaar */
	{
            extern void clear_state();
            clear_state();
            debug_message("Error in reset(), %s\n", ob->name);
	    continue;
        }
	if (d_flag & DEBUG_RESET)
	    fprintf(stderr, "RESET %s\n", ob->name);
	eval_cost = 0; /* Avoid accumulating reset eval /JnA */
	reset_object(ob, 1);
    }
#if TIME_TO_SWAP
    for(ob = obj_list; ob && ob->next_reset < current_time + 5;
	ob = ob->next_all)
    {
	ob->time_of_ref = ob->next_reset; /* mark object for swap-in */
    }
#endif
    memcpy((char *) error_recovery_context,
	   (char *) save_error_recovery_context,
	   sizeof error_recovery_context);
    error_recovery_context_exists = save_rec_exists;
}
    
#if TIME_TO_SWAP
struct object *swap_ob = 0;
look_for_objects_to_swap(int force)
{
    extern long time_to_swap; /* marion - for invocation parameter */
    struct object *ob;
     /*
     * Objects object can be destructed, which means that
     * next object to investigate is saved in next_ob. If very unlucky,
     * that object can be destructed too. In that case, the loop is simply
     * restarted.
     */

    if (interupted && !force ||
	force > 0 && force <= number_swapped)
        return;

    for (ob = (swap_ob ? swap_ob : obj_list); ob; ob = ob->next_all) 
    {
	if (force >= 0)
	{
	    if (ob->time_of_ref > current_time &&
		ob->flags & O_SWAPPED && ob->variables)
		load_ob_from_swap(ob);
	    
	    if (!(ob->flags & (O_HEART_BEAT | O_SWAPPED)) && ob->variables &&
		current_time >= ob->time_of_ref + time_to_swap)
		swap(ob);
	}

	{
	    int i;
	    for (i = 0; i < (int)ob->prog->num_inherited; i++)
	    {
		
		if (ob->prog->inherit[i].prog->line_numbers)
		    swap_lineno(ob->prog->inherit[i].prog);
		
		if (ob->prog->inherit[i].prog->program != (char *)-1 &&
		    current_time >= ob->prog->inherit[i].prog->time_of_ref +
		    time_to_swap)
		    swap_prog(ob->prog->inherit[i].prog);
	    }
	}
	number_swapped++;
	if (interupted && !force ||
	    force > 0 && force <= number_swapped)
	    {
		swap_ob = ob->next_all;
		return;
	    }
    }
    swap_ob = 0;
    number_swapped = 0;
    all_swapped += time_to_swap;
    move_block();
}
#endif

/*
 * Call all heart_beat() functions in all objects.  Also call the next reset,
 * and the call out.
 * We do heart beats by moving each object done to the end of the heart beat
 * list before we call its function, and always using the item at the head
 * of the list as our function to call.  We keep calling heart beats until
 * a timeout or we have done num_heart_objs calls.  It is done this way so
 * that objects can delete heart beating objects from the list from within
 * their heart beat without truncating the current round of heart beats.
 *
 * Set command_giver to current_object if it is a living object. If the object
 * is shadowed, check the shadowed object if living. There is no need to save
 * the value of the command_giver, as the caller resets it to 0 anyway.
 */
static struct object * hb_list = 0; /* head */
static struct object * hb_tail = 0; /* for sane wrap around */

static int num_hb_objs = 0;  /* so we know when to stop! */
static int num_hb_calls = 0; /* stats */
static float perc_hb_probes = 100.0; /* decaying avge of how many complete */

void 
call_heart_beat() 
{
    struct object *ob, *hide_current = current_object;
    int num_done = 0;
    int sum_eval = 0;
    int sum_time = 0;
    int sum_ptime = 0;
    unsigned last_time;
    char hbdesc[100];
    
    comm_time_to_call_heart_beat = 0;
    (void) signal(SIGALRM, catch_alarm);

#ifndef TEST_MODE
    ualarm(1000000 * 2 / HB_PARTS, 1000000 * 2 / HB_PARTS);
#else
    /* This is used when we want to single step in a debugger */
    ualarm(1000000 * 2 / HB_PARTS, 0);
#endif
    last_time = current_time;
    current_time = get_current_time();
    current_interactive = 0;
    if (s_flag)
	reset_mudstatus();
    if (hb_list)
    {
        num_hb_calls++;
	while (hb_list &&
	       !comm_time_to_call_heart_beat
	       && (num_done < num_hb_objs))
	{
            int inh_offset;
	    num_done++;
	    cycle_hb_list();
	    ob = hb_tail; /* now at end */
	    if (!(ob->flags & O_HEART_BEAT))
		fatal("Heart beat not set in object on heart beat list!");
	    if (ob->flags & O_SWAPPED)
		fatal("Heart beat in swapped object.\n");
	    /* move ob to end of list, do ob */
	    if (--ob->hb_count)
		continue;
	    ob->hb_count = ob->hb_rate;
	    
	    if (ob->prog->heart_beat_index == -1)
		continue;
            if (ob->prog->heart_beat_prog == -1)
	    {
		current_prog = ob->prog;
		inh_offset = ob->prog->num_inherited - 1;
	    }
            else
              current_prog =
		  ob->prog->inherit[inh_offset = ob->prog->heart_beat_prog].prog;
	    current_object = ob;
	    current_heart_beat = ob;
	    command_giver = ob;
	    while(command_giver->shadowing)
		command_giver = command_giver->shadowing;
	    if (!(command_giver->flags & O_ENABLE_COMMANDS))
		command_giver = 0;
	    eval_cost = 0;
	    if (s_flag)
		reset_mudstatus();
	    call_function(current_prog, inh_offset,
			  &(current_prog->functions[ob->prog->heart_beat_index]));
	    if (s_flag)
	    {
		sum_eval += eval_cost;
		sum_time += get_millitime();
		sum_ptime += get_processtime();
		sprintf(hbdesc,"HB:%s", ob->name);
		print_mudstatus(hbdesc, eval_cost,
				get_millitime(), get_processtime());
	    }		
	}
	if (num_hb_objs)
	    perc_hb_probes = 100 * (float) num_done / num_hb_objs;
	else
	    perc_hb_probes = 100.0;
    }
    
    current_object = hide_current;
    current_heart_beat = 0;
    command_giver = 0; 		/* Fix by Zappa@Kobra (JnA920503) */
    if (last_time != current_time)
	{
	    look_for_objects_to_reset();
	    if (s_flag)
		{
		    reset_mudstatus();
		    sprintf(hbdesc,"Heartbeat (%d)", num_done);
		    print_mudstatus(hbdesc, sum_eval, sum_time, sum_ptime);
		}
	    call_out();	/* some things depend on this, even without players! */
	}
    
    flush_all_player_mess();
}

/*
 * Take the first object off the heart beat list, place it at the end
 */
static void 
cycle_hb_list()
{
    struct object * ob;
    if (!hb_list)
	fatal("Cycle heart beat list with empty list!");
    if (hb_list == hb_tail)
	return; /* 1 object on list */
    ob = hb_list;
    hb_list = hb_list -> next_heart_beat;
    hb_tail -> next_heart_beat = ob;
    hb_tail = ob;
    ob->next_heart_beat = 0;
}

/*
 * add or remove an object from the heart beat list; does the major check...
 * If an object removes something from the list from within a heart beat,
 * various pointers in call_heart_beat could be stuffed, so we must
 * check current_heart_beat and adjust pointers.
 */

int 
set_heart_beat(struct object *ob, int to)
{
    struct object * o = hb_list;
    struct object * oprev = 0;

    if (ob->flags & O_DESTRUCTED || to < 0)
	return 0;

    while (o && o != ob)
    {
	oprev = o;
	o = o->next_heart_beat;
    }

    if (to)
    {
	if (!(ob->flags & O_HEART_BEAT))
	    {
		ob->flags |= O_HEART_BEAT;
		ob->next_heart_beat = hb_list;
		hb_list = ob;
		if (!hb_tail) hb_tail = ob;
		num_hb_objs++;
		cycle_hb_list();     /* Added by Linus. 911104 */
	    }
	ob->hb_count = to;
	ob->hb_rate = to;
    }
    else 
    {
	if (!(ob->flags & O_HEART_BEAT))
	    return 1;
	
	/* remove all refs */
	ob->flags &= ~O_HEART_BEAT;
	if (hb_list == ob)
	    hb_list = ob->next_heart_beat;
	if (hb_tail == ob)
	    hb_tail = oprev;
	if (oprev)
	    oprev->next_heart_beat = ob->next_heart_beat;
	ob->next_heart_beat = 0;
	num_hb_objs--;
    }

    return(1);
}

/*
 * sigh.  Another status function.
 */
int 
heart_beat_status(int verbose)
{
    char buf[20];

    if (verbose) {
	add_message("\nHeart beat information:\n");
	add_message("-----------------------\n");
	add_message("Number of objects with heart beat: %d, starts: %d\n",
		    num_hb_objs, num_hb_calls);
	sprintf(buf, "%.2f", perc_hb_probes);
	add_message("Percentage of HB calls completed last time: %s\n", buf);
    }
    return 0;
}


/*
 * The epilog() in master.c is supposed to return an array of files to load.
 * The array returned by apply() will be freed at next call of apply(),
 * which means that the ref count has to be incremented to protect against
 * deallocation.
 *
 * The master object is asked to do the actual loading.
 */
void 
preload_objects(int eflag)
{
    struct vector *prefiles;
    struct svalue *ret;
    int ix;

    if (setjmp(error_recovery_context)) 
    {
	clear_state();
	add_message("Error in epilog() in master_ob.\n");
    }
    else
    {
	error_recovery_context_exists = 1;
	push_number(eflag);
	ret = apply_master_ob("epilog", 1);
    }

    if ((ret == 0) || (ret->type != T_POINTER))
	return;
    else
	prefiles = ret->u.vec;

    if ((prefiles == 0) || (prefiles->size < 1))
	return;

    prefiles->ref++; /* Otherwise it will be freed next sapply */

    ix = -1;
    if (setjmp(error_recovery_context)) 
    {
	clear_state();
	add_message("Anomaly in the fabric of world space.\n");
    }
    error_recovery_context_exists = 1;

    while (++ix < prefiles->size) 
    {
	if (prefiles->item[ix].type != T_STRING)
	    continue;
	if (s_flag)
	    reset_mudstatus();
	eval_cost = 0;
	push_string(prefiles->item[ix].u.string, STRING_MALLOC);
	(void)apply_master_ob("preload", 1);
	if (s_flag)
	    print_mudstatus(prefiles->item[ix].u.string, eval_cost,
			    get_millitime(), get_processtime());
#ifdef MALLOC_malloc
	resort_free_list();
#endif
    }
    prefiles->ref--; 
    free_vector(prefiles);
    error_recovery_context_exists = 0;
}

/*
 * catch alarm, set flag for comms code and heart_beat to catch.
 * comms code sets time_to_call_heart_beat for the backend when
 * it has completed the current round of player commands.
 */

void 
catch_alarm(int sig)
{
#if defined(SYSV) || defined(LINUX)
    signal(SIGALRM, catch_alarm);
#endif
    comm_time_to_call_heart_beat = 1;
    interupted = 1;
}

/*
 * All destructed objects are moved int a sperate linked list,
 * and deallocated after program execution.
 */
void 
remove_destructed_objects()
{
    struct object *ob, *next;
    for (ob = obj_list_destruct; ob; ob = next)
    {
	next = ob->next_all;
	destruct2(ob);
    }
    obj_list_destruct = 0;
}

/*
 * Append string to file. Return 0 for failure, otherwise 1.
 */
int 
write_file(char *file, char *str)
{
    FILE *f;

    file = check_valid_path(file, current_object, "write_file", 1);

    if (!file)
	return 0;
    f = fopen(file, "a");
    if (f == 0)
	error("Wrong permissions for opening file %s for append.\n", file);
    if (s_flag)
	num_filewrite++;
    fwrite(str, strlen(str), 1, f);
    fclose(f);
    return 1;
}

int read_file_len; /* Side effect from read_file, so we know how many lines
		      we managed to read */
char *
read_file(char *file, int start, int len)
{
    struct stat st;
    FILE *f;
    char *str, *p, *p2, *end, c;
    int size;

    read_file_len = len;
    if (len < 0) return 0;

    file = check_valid_path(file, current_object, "read_file", 0);

    if (!file)
	return 0;
    f = fopen(file, "r");
    if (f == 0)
	return 0;
    if (fstat(fileno(f), &st) == -1)
	fatal("Could not stat an open file.\n");
    size = st.st_size;
    if (s_flag)
	num_fileread++;
    if (size > READ_FILE_MAX_SIZE)
    {
	if ( start || len )
	    size = READ_FILE_MAX_SIZE;
	else
	{
	    fclose(f);
	    return 0;
	}
    }
    if (!start)
	start = 1;
    if (!len)
	read_file_len = len = READ_FILE_MAX_SIZE;
    str = xalloc(size + 1);
    str[size] = '\0';
    do
    {
	if (size > st.st_size)
	    size = st.st_size;
        if (fread(str, size, 1, f) != 1)
	{
    	    fclose(f);
	    free(str);
    	    return 0;
        }
	st.st_size -= size;
	end = str + size;
        for (p = str; ( p2 = memchr(p, '\n', end - p) ) && --start; )
	    p = p2 + 1;
    } while ( start > 1 );
    
    for (p2 = str; p != end; )
    {
        c = *p++;
	if (!isprint(c) && !isspace(c))
	    c = ' ';
	*p2++ = c;
	if ( c == '\n' )
	    if (!--len)
		break;
    }
    if (len && st.st_size)
    {
	size -= (p2 - str) ; 
	if (size > st.st_size)
	    size = st.st_size;
        if (fread(p2, size, 1, f) != 1)
	{
    	    fclose(f);
	    free(str);
    	    return 0;
        }
	st.st_size -= size;
	end = p2 + size;
        for (; p2 != end; )
	{
	    c = *p2;
	    if (!isprint(c) && !isspace(c))
		*p2 = ' ';
	    p2++;
	    if (c == '\n')
	        if (!--len) break;
	}
	if ( st.st_size && len )
	{
	    /* tried to read more than READ_MAX_FILE_SIZE */
	    fclose(f);
	    free(str);
	    return 0;
	}
    }
    read_file_len -= len;
    *p2 = '\0';
    fclose(f);
    return str;
}


char *
read_bytes(char *file, int start, int len)
{
    struct stat st;

    char *str,*p;
    int size, f;
    int lseek();

    if (len < 0)
	return 0;
    if(len > MAX_BYTE_TRANSFER)
	return 0;

    file = check_valid_path(file, current_object, "read_bytes", 0);

    if (!file)
	return 0;
    f = open(file, O_RDONLY);
    if (f < 0)
	return 0;

    if (fstat(f, &st) == -1)
	fatal("Could not stat an open file.\n");
    size = st.st_size;
    if(start < 0) 
	start = size + start;

    if (start >= size)
    {
	close(f);
	return 0;
    }
    if ((start+len) > size) 
	len = (size - start);

    if ((size = lseek(f,start, 0)) < 0)
    {
	close(f);
	return 0;
    }

    str = xalloc(len + 1);

    size = read(f, str, len);

    close(f);

    if (size <= 0)
    {
	free(str);
	return 0;
    }

    /* We want to allow all characters to pass untouched!
    for (il = 0; il < size; il++) 
	if (!isprint(str[il]) && !isspace(str[il]))
	    str[il] = ' ';

    str[il] = 0;
    */
    /*
     * The string has to end to '\0'!!!
     */
    str[size] = '\0';

    p = string_copy(str);
    free(str);

    return p;
}

int
write_bytes(char *file, int start, char *str)
{
    struct stat st;

    int size, f;
    int lseek();

    file = check_valid_path(file, current_object, "write_bytes", 1);

    if (!file)
	return 0;
    if(strlen(str) > MAX_BYTE_TRANSFER)
	return 0;
    f = open(file, O_WRONLY);
    if (f < 0)
	return 0;

    if (fstat(f, &st) == -1)
	fatal("Could not stat an open file.\n");
    size = st.st_size;
    if(start < 0) 
	start = size + start;

    if (start >= size)
    {
	close(f);
	return 0;
    }
    if ((start+strlen(str)) > size)
    {
	close(f);
	return 0;
    }

    if ((size = lseek(f,start, 0)) < 0)
    {
	close(f);
	return 0;
    }

    size = write(f, str, strlen(str));

    close(f);

    if (size <= 0) {
	return 0;
    }

    return 1;
}


int
file_size(char *file)
{
    struct stat st;

    file = check_valid_path(file, current_object, "file_size", 0);
    if (!file)
	return -1;

    if (file[0] == '/')
	file++;

    if (!legal_path(file))
	return -1;

    if (stat(file, &st) == -1)
	return -1;
    if (S_IFDIR & st.st_mode)
	return -2;
    return st.st_size;
}

int 
file_time(char *file)
{
    struct stat st;

    file = check_valid_path(file, current_object, "file_time", 0);

    if (!file)
	return 0;
    if (stat(file, &st) == -1)
	return 0;
    return st.st_mtime;
}

static double load_av = 0.0;

void 
update_load_av() 
{
    extern double consts[5];
    extern int current_time;
    static int last_time;
    int n;
    double c;
    static int acc = 0;

    acc++;
    if (current_time == last_time)
	return;
    n = current_time - last_time;
    if (n < sizeof consts / sizeof consts[0])
	c = consts[n];
    else
	c = exp(- n / 900.0);
    load_av = c * load_av + acc * (1 - c) / n;
    last_time = current_time;
    acc = 0;
}

static double compile_av = 0.0;

void 
update_compile_av(int lines)
{
    extern double consts[5];
    extern int current_time;
    static int last_time;
    int n;
    double c;
    static int acc = 0;

    acc += lines;
    if (current_time == last_time)
	return;
    n = current_time - last_time;
    if (n < sizeof consts / sizeof consts[0])
	c = consts[n];
    else
	c = exp(- n / 900.0);
    compile_av = c * compile_av + acc * (1 - c) / n;
    last_time = current_time;
    acc = 0;
}

char *
query_load_av()
{
    static char buff[100];

    sprintf(buff, "%.2f cmds/s, %.2f comp lines/s", load_av, compile_av);
    return buff;
}
