
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#ifdef __STDC__
#include <memory.h>
#endif

#include "lint.h"
#include "interpret.h"
#include "object.h"
#include "sent.h"
#include "config.h"
#include "exec.h"
#include "mapping.h"
#include "mudstat.h"

extern int d_flag, s_flag;
extern int total_num_prog_blocks, total_prog_block_size;
extern char *add_slash();
extern void print_svalue (struct svalue *, struct object *);
int tot_alloc_dest_object = 0;
int tot_removed_object = 0;

#ifdef USE_TIOCGETP		/* Check if BSD */
extern int getpid();
#else
extern pid_t getpid();
#endif

#define align(x) ( ((x) + (sizeof(void*)-1) )  &  ~(sizeof(void*)-1) )

extern char *xalloc (int), *string_copy (char *);

void remove_swap_file (struct object *);

extern int atoi();

struct object *previous_ob;
extern struct svalue const0;

int tot_alloc_object, tot_alloc_object_size;

/*
 * Replace carriage return in a string with newlines.
 */

static void 
restore_newline(char *str)
{
    for (; *str; str++)
    {
#ifndef MSDOS
	if (str[0] == '\r')
#else
	if (str[0] == 30)
#endif
	    str[0] = '\n';
    }
}
	
static int failed;
#define Fprintf(s) if (fprintf s == EOF) failed=1

static void save_one (FILE *,struct svalue *);

static void
save_string(FILE *f, char *s)
{
    Fprintf((f, "\""));
    while (*s)
    {
	switch (*s)
	{
	case '"':
	    Fprintf((f, "\\\""));
	    break;
	case '\\':
	    Fprintf((f, "\\\\"));
	    break;
	case '\n':
	    Fprintf((f, "\\n"));
	    break;
	default:
	    /*if (isprint(*s))*/
		Fprintf((f, "%c", *s));
	    break;
	}
	s++;
    }
    Fprintf((f, "\""));
    
}

/*
 * Encode an array of elements.
 */
static void 
save_array(FILE *f, struct vector *v)
{
    int i;

    Fprintf((f, "({"));
    for (i = 0; i < v->size; i++)
    {
	save_one(f, &v->item[i]);
	Fprintf((f,","));
    }
    Fprintf((f,"})"));
}

static void 
save_mapping(FILE *f, struct mapping *m)
{
    int i;
    struct apair *p;

    Fprintf((f, "(["));
    for (i = 0; i < m->size; i++)
    {
	for(p = m->pairs[i]; p; p = p->next)
	{
	    save_one(f, &p->arg);
	    Fprintf((f,":"));
	    save_one(f, &p->val);
	    Fprintf((f,","));
	}
    }
    Fprintf((f,"])"));
}

static void
save_one(FILE *f, struct svalue *v)
{
    switch(v->type) {
    case T_FLOAT:
	Fprintf((f,"#%.8e#", v->u.real));
	break;
    case T_NUMBER:
	Fprintf((f, "%d", v->u.number));
	break;
    case T_STRING:
	save_string(f, v->u.string);
	break;
    case T_POINTER:
	save_array(f, v->u.vec);
	break;
    case T_MAPPING:
	save_mapping(f, v->u.map);
	break;
    case T_OBJECT:
	Fprintf((f,"$%d@%s$", v->u.ob->created,v->u.ob->name));
        break;
    default:
	Fprintf((f, "0"));
	break;
    }
}

/*
 * Save an object to a file.
 * The routine checks with the function "valid_write()" in /obj/master.c
 * to assertain that the write is legal.
 */
void save_object(struct object *ob, char *file)
{
    char *name, tmp_name[80];
    int len, i, j;
    FILE *f;
    int failed = 0;
    /* struct svalue *v; */

    if (ob->flags & O_DESTRUCTED)
	return;

    file = check_valid_path(file, ob, "save_object", 1);
    if (file == 0)
	error("Illegal use of save_object()\n");

    len = strlen(file);
    name = xalloc(len + 3);
    (void)strcpy(name, file);
#ifndef MSDOS
    (void)strcat(name, ".o");
#endif
    /*
     * Write the save-files to different directories, just in case
     * they are on different file systems.
     */
    sprintf(tmp_name, "%s.tmp", name);
#ifdef MSDOS
    (void)strcat(name, ".o");
#endif
    f = fopen(tmp_name, "w");
    if (s_flag)
	num_filewrite++;
    if (f == 0) {
	free(name);
	error("Could not open %s for a save.\n", tmp_name);
    }
    failed = 0;
    
    for (j = 0; j < (int)ob->prog->num_inherited; j++)
    {
	struct program *prog = ob->prog->inherit[j].prog;
	if (ob->prog->inherit[j].type & TYPE_MOD_SECOND)
	    continue;
	for (i = 0; i < (int)prog->num_variables; i++) {
	    struct svalue *v =
		&ob->variables[i + ob->prog->inherit[j].variable_index_offset];
	    
	    if (ob->prog->inherit[j].prog->variable_names[i].type & TYPE_MOD_STATIC)
		continue;
	    if (v->type == T_NUMBER || v->type == T_STRING || v->type == T_POINTER
		|| v->type == T_MAPPING || v->type == T_OBJECT || v->type == T_FLOAT) {
		Fprintf((f, "%s ", ob->prog->inherit[j].prog->variable_names[i].name));
		save_one(f, v);
		Fprintf((f, "\n"));
	    }
	}
    }
    if (fclose(f) == EOF)
	failed = 1;
    if (failed) {
	unlink(tmp_name);
	error("Failed to save to file. Disk could be full.\n");
    }
    (void)unlink(name);
#ifndef MSDOS
    if (link(tmp_name, name) == -1)
#else
    if (rename(tmp_name,name) < 0)
#endif
    {
	perror(name);
	printf("Failed to link %s to %s\n", tmp_name, name);
	error("Failed to save object !\n");
    }
#ifndef MSDOS
    unlink(tmp_name);
#endif
    free(name);
}
void save_map(struct object *ob, struct mapping *map, char *file)
{
    char *name, tmp_name[80];
    struct apair *i;
    int len, j;
    FILE *f;

    int failed = 0;
    /* struct svalue *v; */


    file = check_valid_path(file, ob, "save_object", 1);
    if (file == 0)
	error("Illegal use of save_map()\n");

    len = strlen(file);
    name = xalloc(len + 3);
    (void)strcpy(name, file);
#ifndef MSDOS
    (void)strcat(name, ".o");
#endif
    /*
     * Write the save-files to different directories, just in case
     * they are on different file systems.
     */
    sprintf(tmp_name, "%s.tmp", name);
#ifdef MSDOS
    (void)strcat(name, ".o");
#endif
    f = fopen(tmp_name, "w");
    if (s_flag)
	num_filewrite++;
    if (f == 0) {
	free(name);
	error("Could not open %s for a save.\n", tmp_name);
    }
    failed = 0;
    
    for (j = 0; j < map->size; j++)
    {
	for (i = map->pairs[j]; i; i = i->next) {
	    struct svalue *v =
		&i->val;
	    
	    if (i->arg.type != T_STRING)
		continue;
	    if (v->type == T_NUMBER || v->type == T_STRING || v->type == T_POINTER
		|| v->type == T_MAPPING || v->type == T_OBJECT || v->type == T_FLOAT) {
		Fprintf((f, "%s ", i->arg.u.string));
		save_one(f, v);
		Fprintf((f, "\n"));
	    }
	}
    }
    if (failed)
	error("Failed to save to file. Disk could be full.\n");
    (void)unlink(name);
#ifndef MSDOS
    if (link(tmp_name, name) == -1)
#else
    (void) fclose(f);
    if (rename(tmp_name,name) < 0)
#endif
    {
	perror(name);
	printf("Failed to link %s to %s\n", tmp_name, name);
	error("Failed to save mapping !\n");
    }
#ifndef MSDOS
    (void)fclose(f);
    unlink(tmp_name);
#endif
    free(name);
}

#define BIG 1000

static FILE *resf;
static char *resname;
static int restore_one (struct svalue *, char **);

static struct vector *
restore_array(char **str)
{
    struct svalue *tmp;
    int nmax = BIG;
    int i, k;

    tmp = (struct svalue *)xalloc(nmax * sizeof(struct svalue));
    for(k = 0; k < nmax; k++)
	tmp[k] = const0;
    i = 0;
    for(;;)
    {
	if (**str == '}')
	{
	    if (*++*str == ')')
	    {
		struct vector *v;

		++*str;
		v = allocate_array(i);
		memcpy((char *)&v->item[0], (char *)tmp, sizeof(struct svalue) * i);
		free((char *)tmp);
		return v;
	    }
	    else
		return 0;
	} 
	else
	{
	    if (i > nmax)
	    {
		struct svalue *ntmp;

		nmax *= 2;
		ntmp = (struct svalue *)xalloc(nmax * sizeof(struct svalue));
		memcpy((char *)ntmp, (char *)tmp, sizeof(struct svalue) * nmax);
		free((char *)tmp);
		tmp = ntmp;
		for(k = i; k < nmax; k++)
		    tmp[k] = const0;
	    }
	    if (!restore_one(&tmp[i++], str))
		return 0;
	    if (*(*str)++ != ',')
		return 0;
	}
    }
}

static struct mapping *
restore_mapping(char **str)
{
    struct mapping *m;

    m = allocate_map(0);
    for(;;)
    {
	if (**str == ']')
	{
	    if (*++*str == ')')
	    {
		++*str;
		return m;
	    }
	    else
		return 0;
	}
	else
	{
	    struct svalue arg, val;
	    arg = const0;
	    val = const0;
	    if (!restore_one(&arg, str))
		return 0;
      	    if (*(*str)++ != ':')
		return 0;
	    if (!restore_one(&val, str))
		return 0;
	    assign_svalue(get_map_lvalue(m, &arg, 1), &val);
	    free_svalue(&arg);
	    free_svalue(&val);
	    if (*(*str)++ != ',')
		return 0;
	}
    }
}

static int
restore_one(struct svalue *v, char **sp)
{
    char *q, *p, *s;

    s = *sp;
    switch(*s) {
    case '(':
	switch(*++s)
	{
	case '[':
	    free_svalue(v);
	    v->type = T_MAPPING;
	    s++;
	    v->u.map = restore_mapping(&s);
	    if (!v->u.map) {
		*v = const0;
		(void)fclose(resf);
		error("Illegal mapping format when restoring %s.\n", resname);
	    }
	    break;
	    
	case '{':
	    free_svalue(v);
	    v->type = T_POINTER;
	    s++;
	    v->u.vec = restore_array(&s);
	    if (!v->u.vec) {
		*v = const0;
		(void)fclose(resf);
		error("Illegal array format when restoring %s.\n", resname);
	    }
	    break;
	    
	default:
	    return 0;
	}
	break;
    case '"':
	for(p = s+1, q = s; *p && *p != '"'; p++) 
	{
	    if (*p == '\\') {
		switch (*++p) {
		case 'n':
		    *q++ = '\n';
		    break;
		default:
		    *q++ = *p;
		    break;
		}
	    } else {
		/* Have to be able to restore old format... */
		if (*p == '\r')
		    *q++ = '\n';
		else
		    *q++ = *p;
	    }
	}
	*q = 0;
	if (*p != '"')
            return 0;
	free_svalue(v);
	v->type = T_STRING;
	v->u.string = string_copy(s);
	v->string_type = STRING_MALLOC;
	s = p+1;
	break;
    case '$':
        {
	    int ct;
            struct object *ob;
            char name[1024];
            char *b = strchr(s + 1,'$');

	    *name = '\0';
            if (b == NULL)
                return 0;
            if (sscanf(s,"$%d@%[^$ \n\t]$",&ct,name) != 2)
                return 0;
            ob = find_object2(name);
            if (ob && ob->created == ct)
            {
                free_svalue(v);
                v->type = T_OBJECT;
                v->u.ob = ob;
                add_ref(ob,"restore_one");
            }
            else
            {
                free_svalue(v);
                v->type = T_NUMBER;
                v->u.number = 0;
            }
            s = b + 1;
        }
        break;
    case '#':
	{
	    float f;
	    char *b = strchr(s + 1, '#');
	    if (b == NULL)
		return 0;
	    if (sscanf(s,"#%f#",&f) != 1)
		return 0;
	    free_svalue(v);
	    v->type = T_FLOAT;
	    v->u.real = f;
	    s = b + 1;
	}
	break;
    default:
	if (!isdigit(*s) && *s != '-')
	    return 0;
	free_svalue(v);
	v->type = T_NUMBER;
	v->u.number = atoi(s);
	while(isdigit(*s) || *s == '-')
	    s++;
	break;
    }
    *sp = s;
    return 1;
}

int 
restore_object(struct object *ob, char *file)
{
    char *name, var[100], *val, *buff, *space;
    int len;
    FILE *f;
    struct object *save = current_object;
    struct stat st;
    int p;

    if (current_object != ob)
	fatal("Bad argument to restore_object()\n");
    if (ob->flags & O_DESTRUCTED)
	return 0;

    file = check_valid_path(file, ob, "restore_object", 0);
    if (file == 0)
	error("Illegal use of restore_object()\n");

    len = strlen(file);
    name = xalloc(len + 3);
    (void)strcpy(name, file);
    if (name[len-2] == '.' && name[len-1] == 'c')
	name[len-1] = 'o';
    else
	(void)strcat(name, ".o");
    f = fopen(name, "r");
    if (s_flag)
	num_fileread++;
    if (!f || fstat(fileno(f), &st) == -1) {
	free (name);
	if (f) 
	    (void)fclose(f);
	return 0;
    }
    if (st.st_size == 0) {
	(void)fclose(f);
	free (name);
	return 0;
    }
    val = xalloc(st.st_size + 1);
    buff = xalloc(st.st_size + 1);
    current_object = ob;
    
    while(1) {
	struct svalue *v;

	if (fgets(buff, st.st_size + 1, f) == 0)
	    break;
	/* Remember that we have a newline at end of buff ! */
	space = strchr(buff, ' ');
	if (space == 0 || space - buff >= sizeof (var)) {
	    (void)fclose(f);
	    error("Illegal format when restoring %s.\n", name);
	}
	(void)strncpy(var, buff, space - buff);
	var[space - buff] = '\0';
	p = find_status(ob->prog, var, TYPE_MOD_STATIC);
	if (p == -1)
	    continue;
	v = &ob->variables[p];
	resname = name;
	resf = f;
	space++;
	if (!restore_one(v, &space)) {
	    (void)fclose(f);
	    error("Illegal format when restoring %s.\n", name);
	}
    }
    current_object = save;
    if (d_flag & DEBUG_RESTORE)
	debug_message("Object %s restored from %s.\n", ob->name, name);
    free(name);
    free(buff);
    free(val);
    (void)fclose(f);
    return 1;
}
int 
restore_map(struct object *ob, struct mapping *map, char *file)
{
    char *name, var[100], *val, *buff, *space;
    int len;
    FILE *f;
    struct object *save = current_object;
    struct stat st;
    int p;

    if (current_object != ob)
	fatal("Bad argument to restore_map()\n");
    if (ob->flags & O_DESTRUCTED)
	return 0;

    file = check_valid_path(file, ob, "restore_map", 0);
    if (file == 0)
	error("Illegal use of restore_map()\n");

    len = strlen(file);
    name = xalloc(len + 3);
    (void)strcpy(name, file);
    if (name[len-2] == '.' && name[len-1] == 'c')
	name[len-1] = 'o';
    else
	(void)strcat(name, ".o");
    f = fopen(name, "r");
    if (s_flag)
	num_fileread++;
    if (!f || fstat(fileno(f), &st) == -1) {
	free (name);
	if (f) 
	    (void)fclose(f);
	return 0;
    }
    if (st.st_size == 0) {
	(void)fclose(f);
	free (name);
	return 0;
    }
    val = xalloc(st.st_size + 1);
    buff = xalloc(st.st_size + 1);
    
    while(1) {
	struct svalue v;

	v.type = T_STRING;
	v.string_type = STRING_MALLOC;
	v.u.string = var;

	if (fgets(buff, st.st_size + 1, f) == 0)
	    break;
	/* Remember that we have a newline at end of buff ! */
	space = strchr(buff, ' ');
	if (space == 0 || space - buff >= sizeof (var)) {
	    (void)fclose(f);
	    error("Illegal format when restoring %s.\n", name);
	}
	(void)strncpy(var, buff, space - buff);
	var[space - buff] = '\0';
	
	resname = name;
	resf = f;
	space++;
	if (!restore_one(get_map_lvalue(map,&v,1), &space)) {
	    (void)fclose(f);
	    error("Illegal format when restoring %s.\n", name);
	}
    }
    current_object = save;
    free(name);
    free(buff);
    free(val);
    (void)fclose(f);
    return 1;
}

void 
tell_npc(struct object *ob, char *str)
{
    push_constant_string(str);
    (void)apply("catch_tell", ob, 1, 1);
}

/*
 * Send a message to an object.
 * If it is an interactive object, it will go to his
 * screen. Otherwise, it will go to a local function
 * catch_tell() in that object. This enables communications
 * between players and NPC's, and between other NPC's.
 */
void 
tell_object(struct object *ob, struct svalue *arg)
{
    struct object *save_command_giver;

    if (ob &&(ob->flags & O_DESTRUCTED))
	return;
    print_svalue(arg, ob);
    return;
}

void 
free_object(struct object *ob, char *from)
{
    struct sentence *s;

    ob->ref--;
    if (d_flag & DEBUG_OB_REF)
	printf("Subtr ref to ob %s: %d (%s)\n", ob->name,
		      ob->ref, from);
    if (ob->ref > 0)
	return;
    if (!(ob->flags & O_DESTRUCTED))
    {
	/* This is fatal, and should never happen. */
	fatal("FATAL: Object 0x%x %s ref count 0, but not destructed (from %s).\n",
	    ob, ob->name, from);
    }
    if (ob->interactive)
	fatal("Tried to free an interactive object.\n");
    /*
     * If the program is freed, then we can also free the variable
     * declarations.
     */
    if (ob->name) {
	if (lookup_object_hash(ob->name) == ob)
	    fatal("Freeing object %s but name still in name table", ob->name);
	free(ob->name);
	ob->name = 0;
    }
    tot_alloc_object--;
    tot_alloc_dest_object--;
    tot_removed_object++;
#ifdef RUSAGE
    remove_cpu_stack(ob);
#endif
    free((char *)ob);
    tot_alloc_object_size -= sizeof (struct object);
}

void 
add_ref(struct object *ob, char *from)
{
    ob->ref++;
    if (d_flag & DEBUG_OB_REF)
	printf("Add reference to object %s: %d (%s)\n", ob->name,
	       ob->ref, from);
}

/*
 * Allocate an empty object, and set all variables to 0. Note that a
 * 'struct object' already has space for one variable. So, if no variables
 * are needed, we allocate a space that is smaller than 'struct object'. This
 * unused (last) part must of course (and will not) be referenced.
 */
struct object *
get_empty_object(int num_var)
{
    static struct object NULL_object;
    struct object *ob;
    int size = sizeof (struct object);
    int i;

    tot_alloc_object++;
    tot_alloc_object_size += size;
    ob = (struct object *)xalloc(sizeof(struct object));
    /* marion
     * Don't initialize via memset, this is incorrect. E.g. the bull machines
     * have a (char *)0 which is not zero. We have structure assignment, so
     * use it.
     */
    *ob = NULL_object;
    ob->ref = 1;
    ob->swap_num = 0;
    ob->variables = 0;
    ob->auth = const0;
    return ob;
}

void remove_all_objects()
{
    struct object *ob;
    while(1) {
	if (obj_list == 0)
	    break;
	ob = obj_list;
	destruct_object(ob);
	if (ob == obj_list)
	    break;
    }
    remove_destructed_objects();
}

#if 0
/*
 * For debugging purposes.
 */
void 
check_ob_ref(struct object *ob, char *from)
{
    struct object *o;
    int i;

    for (o = obj_list, i=0; o; o = o->next_all) {
	if (o->inherit == ob)
	    i++;
    }
    if (i+1 > ob->ref) {
	fatal("FATAL too many references to inherited object %s (%d) from %s.\n",
	      ob->name, ob->ref, from);
	if (current_object)
	    fprintf(stderr, "current_object: %s\n", current_object->name);
	for (o = obj_list; o; o = o->next_all) {
	    if (o->inherit != ob)
		continue;
	    fprintf(stderr, "  %s\n", ob->name);
	}
    }
}
#endif /* 0 */

static struct object *hashed_living[LIVING_HASH_SIZE];

static int num_living_names, num_searches = 1, search_length = 1;

static int 
hash_living_name(char *str)
{
#if 1
    return hashstr(str, 100, LIVING_HASH_SIZE);
#else
    unsigned ret = 0;

    while(*str)
	ret = ret * 2 + *str++;
    return ret % LIVING_HASH_SIZE;
#endif
}

struct object *
find_living_object(char *str)
{
    struct object **obp, *tmp;
    struct object **hl;

    num_searches++;
    hl = &hashed_living[hash_living_name(str)];
    for (obp = hl; *obp; obp = &(*obp)->next_hashed_living) {
	search_length++;
	if (!((*obp)->flags & O_ENABLE_COMMANDS))
	    continue;
	if (strcmp((*obp)->living_name, str) == 0)
	    break;
    }
    if (*obp == 0)
	return 0;
    /* Move the found ob first. */
    if (obp == hl)
	return *obp;
    tmp = *obp;
    *obp = tmp->next_hashed_living;
    tmp->next_hashed_living = *hl;
    *hl = tmp;
    return tmp;
}

void 
set_living_name(struct object *ob, char *str)
{
    struct object **hl;

    if (ob->flags & O_DESTRUCTED)
	return;
    if (ob->living_name) {
	remove_living_name(ob);
    }
    if (!*str)
	return;
    num_living_names++;
    hl = &hashed_living[hash_living_name(str)];
    ob->next_hashed_living = *hl;
    *hl = ob;
    ob->living_name = make_shared_string(str);
    return;
}

void 
remove_living_name(struct object *ob)
{
    struct object **hl;

    num_living_names--;
    if (!ob->living_name)
	fatal("remove_living_name: no living name set.\n");
    hl = &hashed_living[hash_living_name(ob->living_name)];
    while(*hl) {
	if (*hl == ob)
	    break;
	hl = &(*hl)->next_hashed_living;
    }
    if (*hl == 0)
	fatal("remove_living_name: Object named %s no in hash list.\n",
	      ob->living_name);
    *hl = ob->next_hashed_living;
    free_string(ob->living_name);
    ob->next_hashed_living = 0;
    ob->living_name = 0;
}

char *
stat_living_objects()
{
    static char tmp[400];

    sprintf(tmp,"Hash table of living objects:\n-----------------------------\n%d living named objects, average search length: %4.2f\n",
	    num_living_names, (double)search_length / num_searches);
    return tmp;
}

void 
reference_prog (struct program *progp, char *from)
{
    progp->ref++;
    if (d_flag & DEBUG_PROG_REF)
	printf("reference_prog: %s ref %d (%s)\n",
	    progp->name, progp->ref, from);
}

/*
 * Decrement reference count for a program. If it is 0, then free the prgram.
 * The flag free_sub_strings tells if the propgram plus all used strings
 * should be freed. They normally are, except when objects are swapped,
 * as we want to be able to read the program in again from the swap area.
 * That means that strings are not swapped.
 */
void 
free_prog(struct program *progp, int free_sub_strings)
{
    int psize;
    extern int total_program_size;
    
    progp->ref--;
    if (d_flag & DEBUG_PROG_REF)
	printf("free_prog: %s\n", progp->name);
    if (progp->ref > 0)
	return;
    if (progp->ref < 0)
	fatal("Negative ref count for prog ref.\n");
    if (progp->program == (char *)-1)
	load_prog_from_swap(progp);
    remove_prog_from_swap(progp);
    
    total_program_size -= (psize = align(progp->program_size) +
			   align(progp->num_strings * sizeof(char *)));
    
    total_prog_block_size -= progp->total_size;
    total_num_prog_blocks--;
    
    if (free_sub_strings) {
	int i;
	
	/* Free all function names. */
	for (i=0; i < (int)progp->num_functions; i++)
	    if (progp->functions[i].name)
		free_string(progp->functions[i].name);
	/* Free all strings */
	for (i=0; i < (int)progp->num_strings; i++)
	    free_string(progp->strings[i]);
	/* Free all variable names */
	for (i=0; i < (int)progp->num_variables; i++)
	    free_string(progp->variable_names[i].name);
	/* Free all inherited objects */
	for (i=0; i < (int)progp->num_inherited - 1; i++)
	{
	    free_prog(progp->inherit[i].prog, 1);
	    free_string(progp->inherit[i].name);
	}
	free_string(progp->inherit[i].name);
	free(progp->name);
    }
    free((char *)progp->program);
    if (progp->line_numbers != 0)
	free((char *)progp->line_numbers);
    free((char *)progp);
}

void 
reset_object(struct object *ob, int arg)
{
    extern int current_time;
#ifdef ns32000
    extern int random_number();
#else
    extern double drand48();
#endif
    extern double log();
    int num_var, i;
    extern int tot_alloc_variable_size;
#if TIME_TO_SWAP > 0
    extern struct object *swap_ob;
#endif
    struct object **pobj;
    double a;
    for(pobj = &obj_list; *pobj && *pobj != ob; pobj = &((*pobj)->next_all));

    if (!*pobj)
	fatal("Object not in the object list!\n");
#if TIME_TO_SWAP > 0
    if (ob == swap_ob)
	swap_ob = ob->next_all;
#endif
    *pobj = ob->next_all;
#ifdef ns32000
    ob->next_reset = current_time + TIME_TO_RESET/2 +
	random_number(TIME_TO_RESET/2, 0);
#else
    a = - log(drand48()) * (float) TIME_TO_RESET;
    if (a > 10000000.0) a = 10000000.0;
    ob->next_reset = current_time + a;
#endif
    for(;*pobj && (*pobj)->next_reset <= ob->next_reset; pobj = &((*pobj)->next_all));
    ob->next_all = *pobj;
    *pobj = ob;

    if (arg == 0)
    {
	if (!(ob->flags & O_CREATED))
	{
	    int num_var;
	    if (num_var = ob->prog->num_variables +
			ob->prog->inherit[ob->prog->num_inherited - 1]
			.variable_index_offset)
	    {
		if (ob->variables)
		    fatal("Object allready initialized!\n");
		
		ob->variables = (struct svalue *)
		    xalloc(num_var * sizeof(struct svalue));
		tot_alloc_variable_size += num_var * sizeof(struct svalue);
		for (i=0; i<num_var; i++)
		    ob->variables[i] = const0;
	    }
	    ob->flags |= O_CREATED;
	    apply(".INIT", ob, 0, 1);
	    apply("create", ob, 0, 1);
	}
    }
    else if (ob->flags & O_CREATED)
    {
        push_number(arg);
	apply("reset", ob, 1, 1);
    }
}

/*
 * If there is a shadow for this object, then the message should be
 * sent to it. But only if catch_tell() is defined. Beware that one of the
 * shadows may be the originator of the message, which means that we must
 * not send the message to that shadow, or any shadows in the linked list
 * before that shadow.
 */
int 
shadow_catch_message(struct object *ob, char *str)
{
    if (!ob->shadowed)
	return 0;
    while(ob->shadowed != 0 && ob->shadowed != current_object)
	ob = ob->shadowed;
    while(ob->shadowing)
    {
	if (function_exists("catch_tell", ob))
	{
	    push_constant_string(str);

	    /* this will work, since we know the function is defined 
	     */
	    if (apply("catch_tell", ob, 1, 1))
		return 1;
	}
	ob = ob->shadowing;
    }
    return 0;
}

/*
 * Returns a list of all inherited files.
 *
 */
struct vector *
inherit_list(struct object *ob)
{
    struct vector *ret;
    struct program *pr;
    char *res;
    int inh;

    ret = allocate_array(ob->prog->num_inherited);

    for (inh = 0; inh < (int)ob->prog->num_inherited; inh++ )
    {
	ret->item[inh].type = T_STRING;
	ret->item[inh].string_type = STRING_SHARED;
	res = add_slash(ob->prog->inherit[inh].prog->name);
	ret->item[inh].u.string = make_shared_string(res);
	free(res);
    }
    return ret;
}

void
change_ref(struct object *to, struct object *from, char *msg)
{
    if (from)
        add_ref(from, msg);
    if (to)
        free_object(to, msg);
}

