
#include <stdio.h>
#include <string.h>

#include "lint.h"
#include "interpret.h"
#include "object.h"
#include "exec.h"
#include "config.h"

struct program *simul_efunp = 0;
struct object *simul_efun_ob = 0;

/* Don't release this pointer ever. It is used elsewhere.
 */
char *simul_efun_file_name;

/*
 * If there is a simul_efun file, then take care of it and extract all
 * information we need.
 */
void 
get_simul_efun()
{
    struct object *ob;
    struct function *funp;
    int i;

#define SIMULFILE "secure/simul_efun.c"

    FILE *f;
    extern struct program *prog;
    extern char *current_file;
    extern int total_lines;
    extern int num_parse_error;
    extern char *inherit_file;
    extern char *current_loaded_file;

    simul_efun_file_name = make_shared_string("/secure/simul_efun");

    f = fopen(SIMULFILE, "r");
    if (f == 0)
    {
	fprintf(stderr, "/secure/simul_efun.c not found.\n");
	exit(1);
    }
#if 1
    start_new_file(f);
    current_file = string_copy(SIMULFILE);	/* This one is freed below */
    current_loaded_file = SIMULFILE;
    compile_file();
    end_new_file();
    current_loaded_file = 0;
    total_lines = 0;
    (void)fclose(f);
    free(current_file);
    current_file = 0;
    
    if (inherit_file || num_parse_error > 0 || prog == 0) 
    {
	fprintf(stderr, "Error in /secure/simul_efun.c.\n");
	return;
    }

    fprintf(stderr,"%s loaded: %d functions\n", SIMULFILE,
	    prog->num_functions);
    if (prog->num_functions == 0)
	return;

    simul_efunp = prog;

    /*
	We make an object so that it can be updated.
     */
    ob = get_empty_object();
    ob->name = string_copy("secure/simul_efun");
    ob->prog = prog;
    ob->next_reset = 0;
    ob->next_all = obj_list;
    obj_list = ob;
    simul_efun_ob = ob;
    add_ref(ob, "simul_efun");
    enter_object_hash(ob);	/* add name to fast object lookup table */
    current_object = 0;
#else
    simul_efun_ob = load_object(simul_efun_file_name, 1);
#endif
    return;
}

/*
 * Reload the simul_efun ob if destructed. 
 * We know that this is possible. Because destruct_object has already
 * tried to recompile it.
 */
void 
assert_simul_efun_ob_loaded()
{
    struct object *save_current = current_object;

    if (!simul_efun_ob || simul_efun_ob->flags & O_DESTRUCTED)
    {
	if (simul_efun_ob)
	    free_object(simul_efun_ob, "simul_efun");

	simul_efun_ob = 0;
	simul_efunp = 0;

	get_simul_efun(); 
	apply_master_ob("simul_efun_reload",0);
	current_object = save_current;
    }
}

/*
   Used by destruct_object to find simul_efun
 */ 
struct object *
query_simul_efun_ob() 
{
    return simul_efun_ob;
}
