#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#ifdef __STDC__
#include <memory.h>
#endif

#include "lint.h"
#include "interpret.h"
#include "object.h"
#include "config.h"
#include "exec.h"
#include "mudstat.h"

#define align(x) ( ((x) + (sizeof(void*)-1) )  &  ~(sizeof(void*)-1) )

#if defined(__NetBSD__) || defined(__bsdi__)
#define	tell(fd)	lseek(fd, 0L, SEEK_CUR)
#endif

#define INTERUPT_SWAP 1
#define TEST_SWAP
/*
 * Swap out programs from objects.
 */

extern int tot_alloc_object_size;
int total_lineno_swapped = 0;
int obj_swapped = 0;
int obj_bytes_swapped = 0;

int num_swapped;
int total_bytes_swapped;
char file_name[100];

int swap_file = -1;		/* The swap file is opened once */

int total_num_prog_blocks = 0, total_prog_block_size = 0;
int prog_code_size = 0, prog_func_size = 0, prog_var_size = 0,
prog_inherit_size = 0, prog_string_size = 0, prog_line_size = 0;
int program_bytes_swapped = 0, total_program_size = 0, programs_swapped = 0;
int last_address = 0, first_hole = 0, allocated_swap = 0, allocated_swap_blocks = 0;
int tot_alloc_variable_size = 0;

int swap_out = 0, swap_in = 0;

extern int interupt;

extern int errno;

extern int d_flag;

int 
assert_swap_file()
{
    if (swap_file == -1) {
#ifndef MSDOS
	char *p, host[50];

	gethostname(host, sizeof host);
	if ((p = strchr(host, '.')) != NULL)
	    *p = '\0';
	sprintf(file_name, "%s.%s.%d", SWAP_FILE, host, getpid());
	swap_file = open(file_name, O_RDWR | O_CREAT,0600);
#else
	swap_file = open(strcpy(file_name,"LPMUD.SWAP"),"w+b");
#endif
	/* Leave this file pointer open ! */
	if (swap_file == -1)
	    return 0;
	write(swap_file, "SWAP", 4);
	last_address = first_hole = tell(swap_file);
    }
    
    return 1;
}


static void
swap_block(int *address, char *block, unsigned int size)
{
    unsigned rsize = align(size + sizeof(char *) + sizeof(int)), osize = 0;
    int *oaddress = 0;
    int nlast_address, chsum, i;
    char *tst;
    
    if (*address == -1)
	return;
    
    if (!assert_swap_file())
	return;

    if (*address == 0)
	{
	    *address = last_address;
	    if (lseek(swap_file, *address, 0) == -1)
		fatal("Couldn't seek the swap file, errno %d, offset %d.\n",
		      errno, *address);
	    if (write(swap_file, (char *)&address, sizeof(long *)) != sizeof(long *))
		{
		    debug_message("I/O error in swap.\n");
		    *address = -1;
		    return;
		}
	    if (write(swap_file, (char *)&rsize, sizeof(unsigned int)) != sizeof(unsigned int))
		{
		    debug_message("I/O error in swap.\n");
		    *address = -1;
		    return;
		}
	    if (write(swap_file, block, size) != size)
		{
		    debug_message("I/O error in swap.\n");
		    *address = -1;
		    return;
		}
	    nlast_address = last_address + rsize;
	    if (first_hole == last_address)
		first_hole = nlast_address;
	    last_address = nlast_address;
	    
	    allocated_swap += rsize;
	    allocated_swap_blocks++;
	    swap_out++;
	    return;
	}
    if (lseek(swap_file, *address, 0) == -1)
	fatal("Couldn't seek the swap file, errno %d, offset %d.\n",
	      errno, *address);
    
    read(swap_file, (char *)&oaddress, sizeof(unsigned *));
    read(swap_file, (char *)&osize, sizeof(int));
    if (oaddress != address)
	fatal("Wrong address in swap file (%d should be %d).\n",
	      oaddress, address);

    if (osize != rsize)
	fatal("Wrong size in swap file (%d should be %d).\n",
	      osize, rsize);

    if (write(swap_file, block, size) != size)
	{
	    debug_message("I/O error in swap.\n");
	    *address = -1;
	    return;
	}
    swap_out++;
    return;
}

static void
delete_block(int *address)
{
    unsigned rsize;
    int *oaddress;
    
    void *i = 0;
    if (lseek(swap_file, *address, 0) == -1)
	fatal("Couldn't seek the swap file, errno %d, offset %d.\n",
	      errno, *address);
    read(swap_file, (char *)&oaddress, sizeof(int *));
    if (oaddress != address)
	fatal("Wrong address in swap file (%d should be %d).\n",
	      oaddress, address);

    if (lseek(swap_file, *address, 0) == -1)
	fatal("Couldn't seek the swap file, errno %d, offset %d.\n",
	      errno, *address);

    if (write(swap_file, (char *)&i, sizeof(int *)) != sizeof(int *))
	debug_message("I/O error in swap.\n");
    read(swap_file, (char *)&rsize, sizeof(unsigned));

    if (*address + rsize == last_address)
	last_address = *address;
    
    if (first_hole > *address)
	first_hole = *address;

    *address = 0;
    allocated_swap -= rsize;
    allocated_swap_blocks--;
    return;
}

void
read_block(int *address, char **block, int size)
{
    int *oaddress = 0, *baddress;
    int i, ochsum,chsum;
    
    unsigned rsize = align(size + sizeof(char *) + sizeof(int)), osize = 0;
    
    if (lseek(swap_file, *address, 0) == -1)
	fatal("Couldn't seek the swap file, errno %d, offset %d.\n",
	      errno, *address);
    
    read(swap_file, (char *)&oaddress, sizeof(unsigned *));
    read(swap_file, (char *)&osize, sizeof(int));
    
    if (oaddress != address)
	fatal("Wrong address in swap file (%d should be %d).\n",
	      oaddress, address);
    if (osize != rsize)
	fatal("Wrong size in swap file (%d should be %d).\n",
	      osize, rsize);

    *block = xalloc(size);
    read(swap_file, *block, size);

    swap_in++;
    
    return;
}

#define BUFFER_SIZE 0x10000
static int block_tail = 0, block_head = 0;
static char *block = 0;

static void
init_block_move(int block_start)
{
    block_tail = block_start;
    block_head = 0;
    block = xalloc(BUFFER_SIZE);
    return;
}

static void
add_block_move(int *address, int rsize)
{
    unsigned size = rsize - sizeof(char *) - sizeof(int);
    *address = block_tail + block_head;
    *(int **)(block + block_head) = address;
    block_head += sizeof(char *);
    *(int *)(block + block_head) = rsize;
    block_head += sizeof(int);
    
    while (size > 0)
    {
	int chunk_size = (BUFFER_SIZE - block_head > size) ? size : (BUFFER_SIZE - block_head);
	read(swap_file, block + block_head, chunk_size);
	size -= chunk_size;
	block_head += chunk_size;
	if (block_head >= BUFFER_SIZE - sizeof(char *) - sizeof(int))
	{
	    int l = tell(swap_file);
	    lseek(swap_file, block_tail, 0);
	    write(swap_file, block, block_head);
	    block_tail += block_head;
	    block_head = 0;
	    lseek(swap_file, l, 0);
	}
    }
    return;
}

static int
flush_block_move(int end)
{
    int r = 0;
    int l;
    void *p = 0;
    
    lseek(swap_file, block_tail, 0);
    write(swap_file, block, block_head);
    l = end - tell(swap_file);
    
    write(swap_file, (char *)&p, sizeof(char *));
    write(swap_file, (char *)&l, sizeof(int));
    free(block);
    r = block_tail + block_head;
    block = 0;
    block_head = block_tail = 0;
    
    return r;
}


int
move_block()
{
    int *address;
    int size;
    int current = first_hole;
    extern struct object *master_ob;
    extern void flush_all_player_mess();
    extern volatile int interupted;
    if (first_hole == last_address)
	return 0;

    init_block_move(first_hole);
    lseek(swap_file, first_hole);

    while(current != last_address)
	{
	    lseek(swap_file, current, 0);
	    read(swap_file, (char *)&address, sizeof(int *));
	    read(swap_file, (char *)&size, sizeof(int));
	    if (address)
		add_block_move(address, size);
	    current += size;
	    if (interupted)
		{
		    first_hole = flush_block_move(current);
		    return 1;
		}
	}
    last_address = first_hole = flush_block_move(current);
}


/*
 * Swap out an object. Only the program is swapped, not the struct object.
 *
 */
int 
swap(struct object *ob)
{
    extern struct object *master_ob;
    extern struct program *simul_efunp;
    extern int errno;
    extern struct svalue *sp, start_of_stack[];
    int size;
    int num_vars, i;
    extern struct svalue const0;
    
    if (ob->flags & (O_DESTRUCTED | O_SWAPPED))
	return 0;
    
    if (d_flag & DEBUG_SWAP)
	{
	    fprintf(stderr, "Swap object %s (ref %d) from 0x%x\n", ob->name, ob->ref, ob->variables);
	}
    if (!ob->variables)
    {
	fprintf(stderr, "warning:no variables in object %s, don't swap it\n",
		ob->name);
	/* Its no good freeing a NULL pointer */
	return 0;
    }

    
    size = (num_vars = ob->prog->inherit[ob->prog->num_inherited - 1].variable_index_offset +
	    ob->prog->num_variables) * sizeof(struct svalue);
    swap_block(&ob->swap_num, (char *)ob->variables, size);
    if (ob->swap_num == -1)
	return;
    free((char *)ob->variables);
    ob->variables = (struct svalue *)-1;
    obj_bytes_swapped += size;
    obj_swapped++;
    tot_alloc_variable_size -= size;

    total_bytes_swapped += size;
    num_swapped++;

    ob->flags |= O_SWAPPED;
    return 1;
}

void 
load_ob_from_swap(struct object *ob)
{
    extern int errno;
    int l, i, size;
    char *vars;
    int num_var;
    struct svalue *baddress;
    

    if (!(ob->flags & O_SWAPPED))
	fatal("Swapping in not swapped out object!\n");

    
    if (d_flag & DEBUG_SWAP)
	{
	    fprintf(stderr,"Unswap object %s (ref %d) from 0x%x\n", ob->name, ob->ref, ob->variables);
	}
    size = (num_var = ob->prog->inherit[ob->prog->num_inherited - 1].variable_index_offset +
	    ob->prog->num_variables) * sizeof(struct svalue);
    
    read_block(&ob->swap_num, (char **)&ob->variables, size);
    
    ob->flags &= ~O_SWAPPED;
    obj_bytes_swapped -= size;
    obj_swapped--;
    tot_alloc_variable_size += size;

    total_bytes_swapped -= size;
    num_swapped--;

}

void 
remove_swap_file(struct object *ob)
{
    /* Haven't implemented this yet :-( */
}

/*
 * This one is called at shutdown. Remove the swap file.
 */

void 
unlink_swap_file() 
{
    if (swap_file == 0)
	return;
#ifndef MSDOS
    unlink(file_name);
    close(swap_file);
#else
    close(swap_file);
    unlink(file_name);
#endif
}

/*
 * Swap out lineno info for a program to the swap file
 * The linenoinfo is separated from the programblock thereby saving memory.
 * This is called directly after compilation from epilog() in postlang.y
 */
int 
swap_lineno(struct program *prog)
{
    int size;
    
    if (d_flag & DEBUG_SWAP)  /* marion */
	debug_message("Swap lineno for %s (ref %d)\n", prog->name);

    size = align(prog->sizeof_line_numbers) +
	align(prog->num_relocs * sizeof(struct reloc));

    if (prog->swap_lineno_index > 0 || !prog->line_numbers) 
    {
        if (prog->line_numbers)
        {
	    total_bytes_swapped += size;
	    total_lineno_swapped += size;
	    num_swapped++;
            free((char*)prog->line_numbers);
	    prog->reloc = 0;
	    prog->line_numbers = 0;
	}
	return 1;
    }

    
    swap_block(&prog->swap_lineno_index, (char *)prog->line_numbers, size);
    if (prog->swap_lineno_index == -1)
	return;
    
    total_lineno_swapped += size;
    total_bytes_swapped += size;
    num_swapped++;
    free((char*)prog->line_numbers);
    prog->line_numbers = 0;
    prog->reloc = 0;
    return 1;
}

/*
  Load the lineno info, this is only done when a runtimeerror occurs
*/
void 
load_lineno_from_swap(struct program *prog)
{
    extern int errno;
    extern int s_flag;
    int size;
    
    if (prog->line_numbers)
	return;
    if (s_flag)
	num_fileread++;

    if (prog->swap_lineno_index == 0)
	fatal("Loading not swapped linenoinfo.\n");
    size = align(prog->sizeof_line_numbers) +
	align(prog->num_relocs * sizeof(struct reloc));
    
    read_block(&prog->swap_lineno_index, (char **)&prog->line_numbers, size);
    prog->reloc = (struct reloc *)((char *)prog->line_numbers + prog->sizeof_line_numbers);

    total_lineno_swapped -= size;
    total_bytes_swapped -= size;
    num_swapped--;

}

int
swap_prog(struct program *prog)
{
    extern int errno;
    int size;

    if (prog->line_numbers)
	swap_lineno(prog);
    if (prog->program == (char *)-1)
	return;
    if (d_flag & DEBUG_SWAP) { /* marion */
	debug_message("Swap program %s (ref %d)\n", prog->name, prog->ref);
    }
    size = align(prog->program_size) +
	align(prog->num_strings * sizeof(char *));
    if (prog->swap_num == 0)
	{
	    swap_block(&prog->swap_num, prog->program, size);
	    if (prog->swap_num == -1)
		fatal("error while swapping program %d.\n", prog->name);
	}
    free(prog->program);
    prog->program = (char *)-1;
    prog->strings = (char **)-1;
    total_program_size -= size;
    program_bytes_swapped += size;
    programs_swapped++;
    total_bytes_swapped += size;
    num_swapped++;
    return 0;
}

load_prog_from_swap(struct program *prog)
{
    extern int errno;
    int l, size;
    
    if (prog->program != (char *)-1 || prog->swap_num == -1)
	return;
    if (d_flag & DEBUG_SWAP) { /* marion */
	debug_message("Unswap program %s (ref %d)\n", prog->name, prog->ref);
    }
    size = align(prog->program_size) +
	align(prog->num_strings * sizeof(char *));
    read_block(&prog->swap_num, &prog->program, size);

    prog->strings = (char **)(prog->program + align(prog->program_size));
    total_program_size += size;
    program_bytes_swapped -= size;
    programs_swapped--;
    total_bytes_swapped -= size;
    num_swapped--;
}

void
remove_ob_from_swap(struct object *ob)
{
    
    if (ob->swap_num > 0)
	{
	    if (ob->flags & O_SWAPPED)
		load_ob_from_swap(ob);
	    
	    delete_block(&ob->swap_num);
	}
    
    
    return;
}

void
remove_prog_from_swap(struct program *prog)
{
    
    if (prog->swap_num > 0)
	{
	    if (prog->program == (char *)-1)
		load_prog_from_swap(prog);
	    
	    delete_block(&prog->swap_num);
	}
    if (prog->swap_lineno_index > 0)
	{
	    if (prog->line_numbers == 0)
		load_lineno_from_swap(prog);
	    delete_block(&prog->swap_lineno_index);
	}
    
    return;
}
